import 'package:flutter/widgets.dart';
import 'package:composable_router/library.dart';

/// Delegates requests from the OS to our Provider and acts on requests from
///   our Provider, also responsible for building the Navigator
class ComposableRouterDelegate extends RouterDelegate<ComposableRoute>
    with ChangeNotifier {
  /// Provider constructor
  ComposableRouterDelegate(
    this._provider, {
    required BuildContext context,
  }) {
    _provider.onRouteChange.listen((event) => notifyListeners());
  }

  /// Responsible for managing our states
  final ComposableRouterProvider _provider;

  /// The navigator key
  final GlobalKey<NavigatorState> _navigator = GlobalKey<NavigatorState>();
  GlobalKey<NavigatorState> get navigatorState => _navigator;
  NavigatorState get navigator => _navigator.currentState!;

  /// Builds the current page stack based on our provider history and current route
  List<Page<dynamic>> buildPageStack(BuildContext context) {
    // Initalize the current stack
    List<Page<dynamic>> pages = [];

    // Build our current route first
    pages.add(_provider.currentRoute.screen(context));

    // Build all our previous routes
    for (int index = _provider.history.length - 1; index >= 0; index--) {
      // Build the route and insert it at the bottom of the array
      ComposableRoute currentPreviousRoute = _provider.history[index];
      pages.insert(0, currentPreviousRoute.screen(context));
    }

    return pages;
  }

  /// Fired when pop is called from the navigator
  bool onPopPage(Route<dynamic> route, dynamic result) {
    return _provider.pop();
  }

  @override
  ComposableRoute get currentConfiguration {
    // We just use our provider state
    return _provider.currentRoute;
  }

  @override
  Future<void> setInitialRoutePath(ComposableRoute route) async {
    // Setup the provider's initial route
    _provider.replace(route);
  }

  @override
  Future<void> setNewRoutePath(ComposableRoute route) async {
    // We handle requests from the os just like any request from the app
    _provider.push(route);
  }

  @override
  Future<bool> popRoute() async {
    // Asks the provider to pop whenever the router is requested to do so
    return _provider.pop();
  }

  @override
  Widget build(BuildContext context) {
    // Builds the router navigator
    return Navigator(
      key: navigatorState,
      pages: buildPageStack(context),
      onPopPage: onPopPage,
    );
  }
}
