import 'dart:async';
import 'dart:collection';
import 'package:flutter/widgets.dart';
import 'package:composable_router/composable_route.dart';

/// The core of the router
class ComposableRouterProvider extends ChangeNotifier {
  ComposableRouterProvider({
    this.historyLimit,
  });

  /// The limit of routes in the history
  int? historyLimit;

  /// Gets the current route
  ComposableRoute get currentRoute => _currentRoute;

  /// Gets the route history
  List<ComposableRoute> get history => UnmodifiableListView(_history);

  /// Gets the allowed routes
  List<ComposableRoute> get allowedRoutes =>
      UnmodifiableListView(_allowedRoutes);

  /// Gets the current build context of the router widget
  BuildContext get context => _context;

  /// Gets the current route change stream
  Stream<ComposableRoute> get onRouteChange =>
      _onCurrentRouteChangeController.stream;

  /// Gets the history change stream
  Stream<List<ComposableRoute>> get onHistoryChange =>
      _onHistoryChangeController.stream;

  /// Gets the allowed routes change stream
  Stream<List<ComposableRoute>> get onAllowedRoutesChange =>
      _onAllowedRoutesChangeController.stream;

  /// On current route change stream controller
  final StreamController<ComposableRoute> _onCurrentRouteChangeController =
      StreamController<ComposableRoute>.broadcast();

  /// On history change stream controller
  final StreamController<List<ComposableRoute>> _onHistoryChangeController =
      StreamController<List<ComposableRoute>>.broadcast();

  /// On allowed routes change stream controller
  final StreamController<List<ComposableRoute>>
      _onAllowedRoutesChangeController =
      StreamController<List<ComposableRoute>>.broadcast();

  /// Initialize the provider
  void initialize(
    BuildContext buildContext, {
    ComposableRoute? initialRoute,
  }) {
    _context = buildContext;
    _currentRoute = initialRoute ?? _currentRoute;
  }

  /// Tries to push the requested route
  bool push(ComposableRoute route) {
    // If we can't enter the route we fail
    if (!_canEnterRoute(route)) return false;

    // Add our current route to our history
    _history.add(_currentRoute);

    // Check if we are over our history limit
    if (historyLimit != null && _history.length > historyLimit!) {
      _history.removeAt(0);
    }

    // Update our current route and history
    _currentRoute = route;

    // Notify the delegates
    _onCurrentRouteChangeController.sink.add(_currentRoute);
    _onHistoryChangeController.sink.add(_history);
    notifyListeners();

    return true;
  }

  /// Tries to replace the current route with the new one
  bool replace(ComposableRoute route) {
    // If we can't enter the route we fail
    if (!_canEnterRoute(route)) return false;

    // Update our current route
    _currentRoute = route;

    // Notify the delegates
    _onCurrentRouteChangeController.sink.add(_currentRoute);
    notifyListeners();

    return true;
  }

  /// Tries to pop our current route
  bool pop() {
    // If our history is empty or we can't enter our previous route we fail
    if (_history.isEmpty || !_canEnterRoute(_history.last)) return false;

    // Update our current route to the last route in history
    _currentRoute = _history.removeLast();

    // Notify the delegates
    _onCurrentRouteChangeController.sink.add(_currentRoute);
    _onHistoryChangeController.sink.add(_history);
    notifyListeners();

    return true;
  }

  /// Tries to pop all the way to the given route
  bool popTo(ComposableRoute route) {
    // While we are not at the target, our history is not empty and we can
    //  enter the previous route
    while (_currentRoute != route && _history.isNotEmpty) {
      // Update our current route to the last route in history
      if (_canEnterRoute(_history.last)) {
        _currentRoute = _history.removeLast();
      }
      // Just pop the last route
      else {
        _history.removeLast();
      }
    }

    // Notify the delegates
    _onCurrentRouteChangeController.sink.add(_currentRoute);
    _onHistoryChangeController.sink.add(_history);
    notifyListeners();

    return true;
  }

  /// Removes all instances of the given route from the history
  void removeFromHistory(ComposableRoute routeToRemove) {
    // Remove from history
    _history.removeWhere((element) => routeToRemove == element);

    // Notify the delegates
    _onHistoryChangeController.sink.add(_history);
    notifyListeners();
  }

  /// Clears the history
  void clearHistory() {
    // We do this so we don't update the stream needlessly
    if (_history.isEmpty) return;

    // Clear the history
    _history.clear();

    // Notify the delegates
    _onHistoryChangeController.sink.add(_history);
    notifyListeners();
  }

  /// Sets the allowed routes to be the given list, after cleaning it
  void setAllowedRoutes(List<ComposableRoute> routes) {
    // Clear the allowed routes
    _allowedRoutes.clear();

    // Fill our allowed routes with all unique routes from the list
    for (ComposableRoute route in routes) {
      if (!_allowedRoutes.contains(route)) {
        _allowedRoutes.add(route);
      }
    }

    // Notify the delegates
    _onAllowedRoutesChangeController.sink.add(_allowedRoutes);
    notifyListeners();
  }

  /// Adds a unique allowed route
  bool addAllowedRoute(ComposableRoute route) {
    // Allowed route list only takes unique values
    if (_allowedRoutes.contains(route)) return false;

    // Add the new route
    _allowedRoutes.add(route);

    // Notify the delegates
    _onAllowedRoutesChangeController.sink.add(_allowedRoutes);
    notifyListeners();

    return true;
  }

  /// Removes the given route from the allowed list
  bool removeAllowedRoute(ComposableRoute route) {
    // The given route is not on the list
    if (!_allowedRoutes.contains(route)) return false;

    // Remove the given route from the allowed list
    bool hasRemoved = _allowedRoutes.remove(route);

    // Notify the delegates if we removed anything
    if (hasRemoved) {
      _onAllowedRoutesChangeController.sink.add(_allowedRoutes);
      notifyListeners();
    }

    return hasRemoved;
  }

  /// Clears the history
  void clearAllowedRoutes() {
    // We do this so we don't update the stream needlessly
    if (_allowedRoutes.isEmpty) return;

    // Clear the allowed routes
    _allowedRoutes.clear();

    // Notify the delegates
    _onAllowedRoutesChangeController.sink.add(_allowedRoutes);
    notifyListeners();
  }

  /// The current route
  late ComposableRoute _currentRoute;

  /// The route history
  List<ComposableRoute> _history = [];

  /// The routes allowed to be taken if any
  List<ComposableRoute> _allowedRoutes = [];

  /// The router build context
  late BuildContext _context;

  /// Check if we can enter the given route
  bool _canEnterRoute(ComposableRoute route) {
    bool isRouteAllowed =
        _allowedRoutes.isEmpty || _allowedRoutes.contains(route);
    bool isRouteRuleMet = route.canEnter(_context);
    return isRouteAllowed && isRouteRuleMet;
  }
}
